package main

import (
	"flag"
	"fmt"
	"html"
	"log"
	"os"
	"strconv"
	"strings"
	"sync"

	"code.rocketnine.space/tslocum/desktop"
	"code.rocketnine.space/tslocum/gmenu/pkg/gmenu"
	"github.com/diamondburned/gotk4/pkg/glib/v2"
	"github.com/diamondburned/gotk4/pkg/gtk/v4"
	"github.com/diamondburned/gotk4/pkg/pango"
	"github.com/kballard/go-shellquote"
)

const (
	iconSize              = 48
	rowMargin             = 4
	labelMarginStart      = 12
	labelMarginTop        = 4
	labelMarginTopComment = 4
)

var execLabel *gtk.Label

func initList(container *gtk.Box) {
	inputView = gtk.NewTextView()
	setNoExpand(&inputView.Widget)
	inputView.SetObjectProperty("accepts-tab", false)
	inputView.SetObjectProperty("wrap-mode", gtk.WrapChar)
	inputView.SetObjectProperty("cursor-visible", false)
	inputView.SetObjectProperty("left-margin", 2)
	inputView.SetObjectProperty("right-margin", 2)

	keyController := gtk.NewEventControllerKey()
	keyController.ConnectKeyPressed(handleKeybinding)
	inputView.AddController(keyController)

	buffer := inputView.Buffer()
	buffer.ConnectChanged(func() {
		gmenu.SetInput(strings.TrimSpace(textViewText(inputView)))
	})

	container.Append(inputView)

	listScroll := gtk.NewScrolledWindow()
	listScroll.SetHExpand(true)
	listScroll.SetVExpand(false)

	listBox = gtk.NewListBox()
	listBox.SetSelectionMode(gtk.SelectionBrowse)
	listBox.SetHExpand(false)
	listBox.SetFocusOnClick(false)

	listScroll.SetPolicy(gtk.PolicyNever, gtk.PolicyAutomatic)
	listScroll.SetChild(listBox)
	listScroll.SetVExpand(true)
	container.Append(listScroll)

	if !config.HideAppIcons {
		d := container.Display()
		iconTheme = gtk.IconThemeGetForDisplay(d)
	}

	lastEntry := len(gmenu.FilteredEntries) - 1

	var (
		entry *gmenu.ListEntry
		wg    = new(sync.WaitGroup)
	)
	for i, label := range gmenu.Names {
		wg.Add(1)

		glib.IdleAdd(rowInitFunc(i, label, entry, lastEntry, wg))
	}

	listBox.SetSortFunc(listSort)
	listBox.SetFilterFunc(listFilter)

	go func() {
		wg.Wait()

		glib.IdleAdd(func() {
			listBox.ConnectRowActivated(func(_ *gtk.ListBoxRow) {
				err := listSelect(inputView, false)
				if err != nil {
					log.Fatal(err)
				}
			})
		})
	}()
}

func initRow(container *gtk.Box, entry *gmenu.ListEntry, i int, lastEntry int) {
	if !config.HideAppIcons {
		img := gtk.NewImage()
		img.SetSizeRequest(iconSize, iconSize)

		container.Append(img)

		glib.IdleAdd(func() { loadIconImage(img, entry) })
	}

	labelContainer := gtk.NewBox(gtk.OrientationVertical, 0)
	labelContainer.SetMarginStart(labelMarginStart)

	labelText := "Shell command"
	if entry != nil && entry.Entry != nil {
		labelText = entry.Label
	}

	l := gtk.NewLabel(fmt.Sprintf("<b>%s</b>", html.EscapeString(labelText)))
	l.SetUseMarkup(true)

	setNoExpand(&l.Widget)
	initLabel(l)

	if !config.HideAppIcons && !config.HideAppDetails {
		l.SetMarginTop(labelMarginTop)
	}

	labelContainer.Append(l)

	if config.HideAppDetails {
		if i == lastEntry {
			execLabel = l
		}
	} else if entry.Entry == nil || (entry.Entry != nil && entry.Comment != "") {
		comment := ""
		if entry.Entry != nil {
			comment = entry.Comment
		}
		l := gtk.NewLabel(comment)
		setNoExpand(&l.Widget)
		initLabel(l)

		if !config.HideAppIcons {
			l.SetMarginTop(labelMarginTopComment)
		}

		labelContainer.Append(l)

		if i == lastEntry {
			execLabel = l
		}
	}

	setNoExpand(&labelContainer.Widget)
	container.Append(labelContainer)

	setNoExpand(&container.Widget)
}

func initLabel(l *gtk.Label) {
	l.SetHAlign(gtk.AlignStart)
	l.SetWrap(false)
	l.SetSingleLineMode(true)
	l.SetEllipsize(pango.EllipsizeEnd)
}

func updateList(input string) {
	execLabel.SetText(input)

	listBox.InvalidateFilter()
	listBox.InvalidateSort()

	listBox.SelectRow(listBox.RowAtIndex(0))
}

func textViewText(tv *gtk.TextView) string {
	buffer := tv.Buffer()
	start, end := buffer.Bounds()

	return buffer.Text(start, end, true)
}

func setNoExpand(v *gtk.Widget) {
	v.SetHExpand(false)
	v.SetVExpand(false)
}

func rowID(row *gtk.ListBoxRow) int {
	if row == nil {
		return -1
	}

	name := row.Name()
	if len(name) < 2 || name[0] != '#' {
		return -1
	}

	id, err := strconv.Atoi(name[1:])
	if err != nil {
		return -1
	}

	return id
}

func selectedIndex() int {
	if listBox == nil {
		return -1
	}

	return rowID(listBox.SelectedRow())
}

func selectedEntry() *desktop.Entry {
	i := selectedIndex()
	if len(gmenu.FilteredEntries) == 0 || i < 0 || i > len(gmenu.FilteredEntries)-1 {
		return nil
	}

	return gmenu.FilteredEntries[i].Entry
}

func listSelect(_ *gtk.TextView, runInTerminal bool) error {
	gmenu.CloseInput()

	var execute string
	entry := selectedEntry()
	if entry == nil {
		execute = textViewText(inputView)
	} else if entry.Type == desktop.Application {
		if entry.Terminal {
			runInTerminal = true
		}

		execute = entry.ExpandExec(shellquote.Join(flag.Args()...))
	} else { // Type == desktop.Link
		execute = shellquote.Join(config.BrowserCommand(), entry.URL)
	}

	path := ""
	if entry != nil {
		path = entry.Path
	}

	err := gmenu.Run(&config.Config, execute, path, runInTerminal, false)
	if err != nil {
		return err
	}

	os.Exit(0)
	return nil
}

func listSort(row1 *gtk.ListBoxRow, row2 *gtk.ListBoxRow) int {
	r1 := rowID(row1)
	r2 := rowID(row2)

	if r1 < 0 {
		if gmenu.MatchEntry(r2) {
			return 1
		}

		return -1
	} else if r2 < 0 {
		if gmenu.MatchEntry(r1) {
			return -1
		}

		return 1
	}

	if gmenu.Sort(r1, r2) {
		return -1
	}

	return 1
}

func listFilter(row *gtk.ListBoxRow) bool {
	match := gmenu.MatchEntry(rowID(row))

	row.SetSelectable(match)

	return match
}

func rowInitFunc(i int, label string, entry *gmenu.ListEntry, lastEntry int, wg *sync.WaitGroup) func() {
	return func() {
		row := gtk.NewListBoxRow()

		if i == lastEntry {
			entry = &gmenu.ListEntry{Entry: nil, Label: ""}
		} else {
			row.SetName("#" + strconv.Itoa(i))

			entry = &gmenu.ListEntry{Entry: gmenu.Entries[i], Label: label}
		}

		container := gtk.NewBox(gtk.OrientationHorizontal, 0)
		container.SetMarginStart(rowMargin)
		container.SetMarginTop(rowMargin)
		container.SetMarginBottom(rowMargin)
		container.SetMarginEnd(rowMargin)

		initRow(container, entry, i, lastEntry)

		row.SetChild(container)
		listBox.Append(row)

		listBox.SelectRow(listBox.RowAtIndex(0))
		row.Show()

		wg.Done()
	}
}
